#include <vector>
#include <span>
#include <thread>
#include <mutex>

template<typename T, typename C>
class MergeSorterMT {

public:
    MergeSorterMT(C cmp) : cmp(cmp){}


    C cmp;
    std::recursive_mutex mut;

    auto merge(std::span<T> &output, std::span<T> left, std::span<T> right) -> void {
        std::vector<T> buf;
        buf.reserve(left.size() + right.size());

        auto l = left.begin();
        auto r = right.begin();
        auto o = buf.begin();

        while (l < left.end() && r < right.end()) {
            if (cmp(*l, *r)) {
                buf.insert(o, *l);
                l++;
            } else {
                buf.insert(o, *r);
                r++;
            }
            o++;
        }
        while (l < left.end()) {
            buf.insert(o, *l);
            o++;
            l++;
        }
        while (r < right.end()) {
            buf.insert(o, *r);
            o++;
            r++;
        }

        {
            std::lock_guard<std::recursive_mutex> lock(mut);
            std::move(buf.begin(), buf.end(), output.begin());
        }
    }

    auto split(std::span<T> &data, int depth, int &max_depth, std::recursive_mutex &mut) -> void {
        if(std::distance(data.begin(), data.end()) <= 1) {
            return;
        }

        auto mid = data.begin();
        std::advance(mid, std::distance(data.begin(), data.end())/2);

        std::span<T> left(data.begin(), mid);
        std::span<T> right(mid, data.end());

        if(depth < max_depth) {
            std::thread left_thread(&MergeSorterMT<T, C>::split, this, left, depth + 1, max_depth, mut);
            std::thread right_thread(&MergeSorterMT<T, C>::split, this, right, depth + 1, max_depth, mut);

            left_thread.join();
            right_thread.join();
        } else {
            split(left, depth + 1, max_depth, mut);
            split(right, depth + 1, max_depth, mut);
        }

        merge(data, left, right);
    }




};